from collections import deque
from collections.abc import Sequence

from .tokens import ClosingTag, Content, OpeningTag, Token
from .types import Element, HtmlChild


class Tokenizer:
    def __init__(self) -> None:
        self.nodes: deque[HtmlChild] = deque()
        self.tokens: deque[Token] = deque()
        self.open_nodes: set[HtmlChild] = set()

    def tokenize(self, root_nodes: Sequence[Element]) -> deque[Token]:
        self.add_children(root_nodes)

        while self.nodes:
            node = self.nodes[-1]

            if isinstance(node, str):
                self.nodes.pop()
                self.add_token(Content(node))

            else:  # Node
                if node not in self.open_nodes:
                    if node.self_closing:
                        # We don't want to come back and process this node again.
                        self.nodes.pop()

                    self.add_children(node.children)
                    self.add_token(OpeningTag(node.tag, node.attrs))

                else:  # Closing
                    self.nodes.pop()
                    self.add_token(ClosingTag(node.tag))
                    self.open_nodes.remove(node)

                self.open_nodes.add(node)

        return self.tokens

    def add_children(self, children: Sequence[HtmlChild]) -> None:
        self.nodes.extend(reversed(children))

    def add_token(self, token: Token) -> None:
        self.tokens.append(token)



from collections.abc import Sequence
from typing import Union

from .utils import is_self_closing_tag

HtmlAttributes = dict[str, object]
HtmlChild = Union["Element", str]
HtmlChildren = Sequence[HtmlChild] | HtmlChild


class Element:
    def __init__(
        self,
        tag: str,
        attrs: HtmlAttributes | None = None,
        children: list[HtmlChild] | None = None,
    ):
        self.tag = tag
        self.attrs = attrs or {}
        self.children = children or []

    @property
    def self_closing(self) -> bool:
        return is_self_closing_tag(self.tag)


class SafeString(str): ...



from collections import deque
from html import escape
from typing import TYPE_CHECKING

from .tokens import ClosingTag, Content, OpeningTag, Token
from .types import SafeString

if TYPE_CHECKING:
    from .types import HtmlAttributes


class Compiler:
    def compile(self, tokens: deque[Token]) -> str:
        self.tokens = tokens

        self.depth: int = 0
        self.token: Token | None = tokens.popleft()
        self.code: deque[str] = deque()

        while self.token:
            getattr(self, f"visit_{self.token.__class__.__name__}")(self.token)
            self.move_cursor(self.token, self.peek())
            self.eat()

        if self.code[-1] != "\n":
            self.newline()

        return "".join(self.code)

    def eat(self) -> None:
        try:
            self.token = self.tokens.popleft()
        except IndexError:
            self.token = None

    def peek(self) -> Token | None:
        try:
            return self.tokens[0]
        except IndexError:
            return None

    def append(self, fragment: str) -> None:
        self.code.append(fragment)

    def indent(self) -> None:
        self.append("    " * self.depth)

    def newline(self) -> None:
        self.append("\n")

    def visit_OpeningTag(self, tag: OpeningTag) -> None:
        attrs = self.render_attrs(tag.attrs)
        attrs = " " + attrs if attrs else ""
        self.append(f"<{tag.name}{attrs}>")

    def visit_Content(self, content: Content) -> None:
        string = (
            content.string
            if isinstance(content.string, SafeString)
            else escape(content.string)
        )
        self.append(string)

    def visit_ClosingTag(self, tag: ClosingTag) -> None:
        self.append(f"</{tag.name}>")

    def move_cursor(self, token: Token, next_token: Token | None) -> None:
        if not next_token:
            return

        match (token, next_token):
            case [OpeningTag(is_self_closing=False), ClosingTag()]:
                self.depth = self.depth
            case [OpeningTag(is_self_closing=False), _]:
                self.depth += 1
            case [_, ClosingTag()]:
                self.depth -= 1

        match (token, next_token):
            # current or next token is an opening block tag
            case [OpeningTag(is_block=True), _] | [_, OpeningTag(is_block=True)]:
                self.newline()
                self.indent()
            # current or next token is a closing block tag
            case [ClosingTag(is_block=True), _] | [_, ClosingTag(is_block=True)]:
                self.newline()
                self.indent()

    @classmethod
    def render_attrs(cls, attrs: "HtmlAttributes") -> str:
        if not attrs:
            return ""
        attrs_list = [cls.render_attr(k, v) for k, v in attrs.items()]
        return " ".join(attrs_list)

    @classmethod
    def render_attr(cls, key: str, value: object) -> str:
        if value == "":
            return key

        if value is True:
            return key

        if value is False:
            return ""

        if key == "style":
            if not isinstance(value, dict):
                raise ValueError("A style value must be a dict")

            return f'{key}="{cls.render_style(value)}"'
        # Default
        return f'{key}="{value}"'

    @classmethod
    def render_style(cls, style: dict[str, str]) -> str:
        if not style:
            return ""
        style_list = [f"{k}: {v}" for k, v in style.items()]
        return "; ".join(style_list)



from .types import HtmlAttributes
from .utils import is_block_tag, is_self_closing_tag


class Token:
    pass


# TODO: good use case for dataclasses?
class OpeningTag(Token):
    def __init__(self, name: str, attrs: HtmlAttributes):
        self.name = name
        self.attrs = attrs

    @property
    def is_block(self) -> bool:
        return is_block_tag(self.name)

    @property
    def is_self_closing(self) -> bool:
        return is_self_closing_tag(self.name)

    def __repr__(self) -> str:
        return f"OpeningTag({self.name!r}, {self.attrs!r})"


class Content(Token):
    def __init__(self, string: str):
        self.string = string

    def __repr__(self) -> str:
        return f"Content({self.string!r})"


class ClosingTag(Token):
    def __init__(self, name: str):
        self.name = name

    @property
    def is_block(self) -> bool:
        return is_block_tag(self.name)

    def __repr__(self) -> str:
        return f"ClosingTag({self.name!r})"



from collections.abc import Sequence
from typing import overload

from .compiler import Compiler
from .tokenizer import Tokenizer
from .types import Element, HtmlAttributes, HtmlChild, HtmlChildren, SafeString

__all__ = ["h", "render", "safe", "Element", "SafeString"]


@overload
def h(
    tag: str,
    /,
) -> Element: ...  # pragma: no cover


@overload
def h(
    tag: str,
    attrs: HtmlAttributes,
    /,
) -> Element: ...  # pragma: no cover


@overload
def h(
    tag: str,
    children: HtmlChildren,
    /,
) -> Element: ...  # pragma: no cover


@overload
def h(
    tag: str,
    attrs: HtmlAttributes,
    children: HtmlChildren,
    /,
) -> Element: ...  # pragma: no cover


def h(
    tag: str,
    attrs_or_children: HtmlAttributes | HtmlChildren | None = None,
    children: HtmlChildren | None = None,
    /,
) -> Element:
    """
    Used for building html using python functions.

    1 arg - tag
    2 args - tag, attrs or children
    3 args - tag, attrs, and children
    """
    tag, attrs, children = _handle_args(tag, attrs_or_children, children)
    return Element(tag, attrs, children)


def render(elements: Element | Sequence[Element]) -> str:
    if isinstance(elements, Element):
        elements = [elements]

    tokens = Tokenizer().tokenize(elements)
    return Compiler().compile(tokens)


def safe(string: str) -> SafeString:
    return SafeString(string)


def _handle_args(
    tag: str,
    attrs_or_children: HtmlAttributes | HtmlChildren | None = None,
    children: HtmlChildren | None = None,
    /,
) -> tuple[str, HtmlAttributes, list[HtmlChild]]:
    args = (tag, attrs_or_children, children)
    match args:
        # 1: h("")
        case [str(), None, None]:
            return args[0], {}, []
        # 2: h("", {})
        case [str(), dict(), None]:
            return args[0], args[1], []
        # 2: h("", [])
        case [str(), list(), None]:
            return args[0], {}, args[1]
        # 2: h("", h("")) OR h("", "")
        case [str(), Element() | str(), None]:
            return args[0], {}, [args[1]]
        # 3: h("", {}, [])
        case [str(), dict(), list()]:
            return args[0], args[1], args[2]
        # 3: h("", {}, h("")) OR h("", {}, "")
        case [str(), dict(), Element() | str()]:
            return args[0], args[1], [args[2]]
        case _:
            raise ValueError("Invalid arguments")



from .constants import block_tags, self_closing_tags


def is_self_closing_tag(tag: str) -> bool:
    return tag in self_closing_tags


def is_block_tag(tag: str) -> bool:
    return tag in block_tags
